!
! Module: esky2det_mod
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: This task transforms sky coordinates to either DETX/Y or RAWX/Y.
! 
! Comments:
!	TODO:
!	- Write units and comment to any new columns.
!	- Overwrite old INSTRUME kwd??
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module esky2det_mod

  use dal
  use param  
  use errorHandling
  use cal		! -> BoresightT
  use caloaldefs

!  use caloalutils ! ???? doesn't this have the instId functions? But esky2det seems to work without it! Maybe there is some link via the cal.

  use cal_aux	! for getCalInfo
  use constants
  use dal_aux	! for readColDataToFixed, splitSetTabName, addOrOpenColumn.
  use ssc_misc	! for intToStr
  use coordinate	! for raDecToInst and instToRaw;
			! also for detUnitArcsec definition.
!use geometric_types !*** temporary

  implicit none 

  ! Command-line parameters:

  logical ::&
    withUncertCol = .false.,&
    withUncertRadius = .false.,&
    withHeader = .false.,&
    withOutSet = .false.
  character(255) ::&
    inSetTabName = '',&
    outSetTabName = '',&
    calInfoSetName = ''
  character(20) ::&
    dataStyle = '',&
    outputUnit = '',&
    userInstrument = '',&
    mosNodeStr = '',&
    calInfoStyle = '',&
    userDateTimeStr = ''
  real(single) ::&
    userRaDeg = 0.0,&
    userDecDeg = 0.0,&
    userSkyUncertRadiusDeg = 0.0,&
    userScAttitudeRaDeg = 0.0,&
    userScAttitudeDecDeg = 0.0,&
    userScAttitudeAposDeg = 0.0

  character(255) ::&
    inSetName = '',&
    outSetName = '',&
    inTabName = '',&
    outTabName = ''
  character(20) ::&
    raColName = '',&
    decColName = '',&
    raUncertColName = '',&
    decUncertColName = '',&
    raDecUncertColName = ''
  character(20) ::&
    rawXcolName = '',&
    rawYcolName = '',&
    rawXUncertColName = '',&
    rawYUncertColName = '',&
    ccdNrColName = '',&
    detXcolName = '',&
    detYcolName = '',&
    detUncertColName = ''
  character(20), parameter :: myDefaultStr = 'null_default'

  integer :: OFF_CHIP_BIT = 0

  private 

  public esky2det

contains 

  !---------------------------------------------------------------------
  subroutine esky2det
  !---------------------------------------------------------------------

    implicit none

    real(double), dimension(:), pointer :: raDeg, decDeg,&
      skyUncertRadiusDeg
!    real(single),   dimension(:), allocatable :: detX, detY, detUncert
real(double),   dimension(:), allocatable :: detX, detY, detUncert
!    integer(int16), dimension(:), allocatable ::&
!real(single), dimension(:), allocatable ::&
real(double), dimension(:), allocatable ::&
      rawX, rawY, rawXUncert, rawYUncert
    integer(int16) :: vsize
    integer(int32) :: instrumentId
    real(double) :: localTimeStamp
    type(SpacecraftAttitudeType) :: scAttitude, instrumAttitude
    real(double) ::&
      focalLengthMm,&
      pixelWidthXMm,&
      pixelWidthYMm
    logical(bool), allocatable :: offChip(:)
    integer(int8), allocatable :: ccdNum(:)
    integer :: mosNode

    !-------------------------------------------------------------------
    ! Start

    call readParameters

    !...................................................................
    ! Get the input data:

    call readRaDec(raDeg, decDeg, skyUncertRadiusDeg)

    !...................................................................
    ! Cal setups:

    select case(calInfoStyle)
    case('set')
      call getCalInfo(calInfoSetName, instrumentId, localTimeStamp&
        , scAttitude, instrumAttitude) ! in ssclib/cal_aux. This also
      ! sets the instrument state of the cal.

    case('user')
      instrumentId = instId(userInstrument)
      localTimeStamp = toTimeTag(userDateTimeStr)
      ! in caloalutils/caloaldefs.f90

      scAttitude%ra   = userScAttitudeRaDeg   * PiOn180
      scAttitude%dec  = userScAttitudeDecDeg  * PiOn180
      scAttitude%apos = userScAttitudeAposDeg * PiOn180

      call CAL_setState(instrument=instrumentId, randomize=.false.)

    case default
call error('badCalInfoStyle', "The value of the --calinfostyle &
&parameter was not recognised.")
    end select

!    write(errstr, '(a, f12.2, a)') 'Timestamp = ', myTimeStamp, ' sec.'
!    call message(AppMsg, Verbose)

    !...................................................................
    ! Now do the actual transformation (then output):

    vsize = size(raDeg)
    allocate(detX(vsize), detY(vsize))

!    call raDecToInst(raDeg, decDeg, scAttitude&
!      , localTimeStamp, detX, detY)
call raDecToDet(raDeg*PiOn180, decDeg*PiOn180, scAttitude&
      , localTimeStamp, detX, detY)

    select case(outputUnit)
    case('det')
      allocate(detUncert(vsize))
      detUncert = skyUncertRadiusDeg * 3600.0 / detUnitArcsec
      call detOutput(raDeg, decDeg, detX, detY, detUncert, instrumentId)
      deallocate(detUncert)

    case('raw')
      allocate(rawX(vsize), rawY(vsize), rawXUncert(vsize)&
        , rawYUncert(vsize), ccdNum(vsize), offChip(vsize))

      if (instrumentId .eq. EMOS1 .or. instrumentId .eq. EMOS2) then
        if (mosNodeStr .eq. 'primary') then
          mosNode = EMOS_PRIMARY_NODE
        else if (mosNodeStr .eq. 'redundant') then
          mosNode = EMOS_REDUNDANT_NODE
        else
call error('badNodeStr', 'The mos node string '//trim(mosNodeStr)&
//' was not recognised.')
        end if
        call CAL_setState(ccdNodeId=mosNode)
      end if

!      call instToRaw(rawX, rawY, offChip, ccdNum, detX, detY)
!!      call l_instToRawVectorReal32Ccd(rawX, rawY, offChip, ccdNum, detX, detY)
call detToRaw(detX, detY, rawX, rawY, offChip, ccdNum) ! in ssclib/coordinate.

      call CAL_getMiscellaneousData('FOCAL_LENGTH', focalLengthMm)
      call CAL_getMiscellaneousData('MM_PER_PIXEL_X', pixelWidthXMm)
      call CAL_getMiscellaneousData('MM_PER_PIXEL_Y', pixelWidthYMm)
!      rawXUncert = nint(skyUncertRadiusDeg * PiOn180 * focalLengthMm&
!        / pixelWidthXMm)
!      rawYUncert = nint(skyUncertRadiusDeg * PiOn180 * focalLengthMm&
!        / pixelWidthYMm)
rawXUncert = skyUncertRadiusDeg * PiOn180 * focalLengthMm&
  / pixelWidthXMm
rawYUncert = skyUncertRadiusDeg * PiOn180 * focalLengthMm&
  / pixelWidthYMm

      call rawOutput(raDeg, decDeg, real(rawX, single), real(rawY, single)&
        , real(rawXUncert, single), real(rawYUncert, single)&
        , offChip, ccdNum, instrumentId)

      deallocate(rawX, rawY, rawXUncert, rawYUncert, offChip, ccdNum)

    case default
call error('badOutUnit', "The value of the --outunit parameter &
&was not recognised.")
    end select

    deallocate(raDeg, decDeg, skyUncertRadiusDeg)
    deallocate(detX, detY)

  end subroutine esky2det

  !---------------------------------------------------------------------
  subroutine readParameters
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start:

    call message(AppMsg, Verbose, 'Reading parameters.')

    dataStyle = stringParameter('datastyle')
    select case(dataStyle)
    case('set')
      inSetTabName = stringParameter('intab')
      raColName          = 'RA'
      decColName         = 'DEC'
!***** move these initialisations to the header.

      withUncertCol = booleanParameter('witherrorcol')
      if (withUncertCol) then
        raUncertColName    = 'RA_ERR'
        decUncertColName   = 'DEC_ERR'
        raDecUncertColName = 'RADEC_ERR'
!***** move these initialisations to the header.
      end if

      withOutSet = booleanParameter('withouttab')
      outSetTabName = stringParameter('outtab')
      rawXcolName        = 'RAWX'
      rawYcolName        = 'RAWY'
      rawXUncertColName  = 'RAWX_ERR'
      rawYUncertColName  = 'RAWY_ERR'
      ccdNrColName       = 'CCDNR'
      detXcolName        = 'DETX'
      detYcolName        = 'DETY'
      detUncertColName   = 'DET_ERR'
!***** move these initialisations to the header.

    case('user')
      userRaDeg = realParameter('ra')
      userDecDeg = realParameter('dec')

      withUncertRadius = booleanParameter('witherrorradius')
      if (withUncertRadius) then
        userSkyUncertRadiusDeg = realParameter('errorradius')
      end if

      withHeader = booleanParameter('withheader')

    case default
call error('badDataStyle',"The value of the --datastyle parameter &
&was not recognised.")
    end select

    outputUnit = stringParameter('outunit')
    select case(outputUnit)
    case('det')
    case('raw')
      mosNodeStr = stringParameter('mosccdnode')
    case default
call error('badOutUnit',"The value of the --outunit parameter &
&was not recognised.")
    end select

    calInfoStyle = stringParameter('calinfostyle')
    select case(calInfoStyle)
    case('set')
      calInfoSetName = stringParameter('calinfoset')

    case('user')
      userInstrument = stringParameter('instrument')
      userDateTimeStr = stringParameter('datetime')
      userScAttitudeRaDeg = realParameter('scattra')
      userScAttitudeDecDeg = realParameter('scattdec')
      userScAttitudeAposDeg = realParameter('scattapos')

    case default
call error('badCalInfoStyle',"The value of the --calinfostyle &
&parameter was not recognised.")
    end select

  end subroutine readParameters

  !---------------------------------------------------------------------
  subroutine readRaDec(raDeg, decDeg, skyUncertRadiusDeg)
  !---------------------------------------------------------------------

    implicit none

    real(double), dimension(:), pointer :: raDeg, decDeg,&
      skyUncertRadiusDeg

    real(double), dimension(:), allocatable :: trueRaUncertArcmin, raUncertArcsec, decUncertArcsec, raDecUncertArcsec
    integer(int16) :: vsize
    type(DataSetT) :: set
    type(TableT) :: tab

    !-------------------------------------------------------------------
    ! Start

    call message(AppMsg, Verbose, 'Reading input data.')

    select case(dataStyle)
    case('set')
      call splitSetTabName(inSetTabName, inSetName, inTabName) ! in ssclib/dal_aux

      set = dataSet(inSetName, READ)
      tab = table(set, inTabName)
      vsize = numberOfRows(tab)
      allocate(raDeg(vsize), decDeg(vsize))
      call readColDataToFixed(tab,  raColName, raDeg)
      call readColDataToFixed(tab, decColName, decDeg)
!      vsize = size(raDeg)
      allocate(skyUncertRadiusDeg(vsize))

      if (withUncertCol) then
        if (hasColumn(tab, raUncertColName)&
        .and. hasColumn(tab, decUncertColName)) then
          allocate(raUncertArcsec(vsize), decUncertArcsec(vsize))
          call readColDataToFixed(tab,  raUncertColName, raUncertArcsec)
          call readColDataToFixed(tab, decUncertColName, decUncertArcsec)

          allocate(trueRaUncertArcmin(vsize))

          trueRaUncertArcmin = raUncertArcsec * cos(decDeg * piOn180)
          where(trueRaUncertArcmin .gt. decUncertArcsec)
            skyUncertRadiusDeg  = raUncertArcsec  / 3600.0
          elsewhere
            skyUncertRadiusDeg = decUncertArcsec / 3600.0
          end where
          deallocate(raUncertArcsec, decUncertArcsec, trueRaUncertArcmin)

        else
          allocate(raDecUncertArcsec(vsize))
          call readColDataToFixed(tab,  raDecUncertColName, raDecUncertArcsec)
!          vsize = size(raDecUncertArcsec)
          skyUncertRadiusDeg = raDecUncertArcsec / 3600.0
          deallocate(raDecUncertArcsec)
        end if
      end if

      call release(set)

    case('user')
      allocate(raDeg(1), decDeg(1), skyUncertRadiusDeg(1))
      raDeg(1)              = userRaDeg
      decDeg(1)             = userDecDeg
      skyUncertRadiusDeg(1) = userSkyUncertRadiusDeg

    case default
call error('badDataStyle', "The value of the --datastyle parameter was &
&not recognised.")
    end select

  end subroutine readRaDec

  !---------------------------------------------------------------------
  subroutine detOutput(raDeg, decDeg, detX, detY, detUncerts&
  , instrumentId)
  !---------------------------------------------------------------------

    implicit none

    real(double), intent(in) :: raDeg(:),&
                                decDeg(size(raDeg))
!    real(single), intent(in) :: detX(size(raDeg)),&
real(double), intent(in) :: detX(size(raDeg)),&
                                detY(size(raDeg)),&
                                detUncerts(size(raDeg))
    integer(int32), intent(in) :: instrumentId

    type(DataSetT) :: set
    type(TableT)   :: tab
    type(ColumnT)  :: detXcol, detYcol, detUncertsCol
    real(single), dimension(:), pointer :: dalDetX, dalDetY, dalDetUncerts
    character(37) :: outputStr
    integer(int16) :: i
    real(double) ::&
      fovRadiusDeg, fovRadiusDet, fovRadiusDet2
    character(12) :: srcNumStr

    !-------------------------------------------------------------------
    ! Start

    call message(AppMsg, Verbose, 'Writing output data.')

    call CAL_getMiscellaneousData('FOV_RADIUS', fovRadiusDeg)
    fovRadiusDet = fovRadiusDeg * 3600.0 / detUnitArcsec ! roughly
    ! speaking. Have to take tan for exact value.
    fovRadiusDet2 = fovRadiusDet * fovRadiusDet

    select case(dataStyle)
    case('set')
      if (withOutSet) then
        call splitSetTabName(outSetTabName, outSetName, outTabName)
        ! in ssclib/dal_aux

if (outTabName .eq. '') call error('emptyTableName',"Parameter --outtab &
&should be given in the form dataSetName:tableName")

        set = dataSet(outSetName, CREATE)
        call setAttribute(set, 'INSTRUME', trim(instId(instrumentId)))
!*** put in DATE-TIME and wcs and *_PNT kwds too??
        tab = addTable(set, outTabName, size(detX))

      else
        set = dataSet(inSetName, MODIFY)
        tab = table(set, inTabName, MODIFY)
      end if

      detXcol = addOrOpenColumn(tab, 'DETX', REAL32)
      detYcol = addOrOpenColumn(tab, 'DETY', REAL32)
      dalDetX  => real32Data(detXcol)
      dalDetY  => real32Data(detYcol)
      dalDetX  = real(detX, single)
      dalDetY  = real(detY, single)

      if (withUncertCol) then
        detUncertsCol = addOrOpenColumn(tab, 'DET_ERR', REAL32)
        dalDetUncerts  => real32Data(detUncertsCol)
        dalDetUncerts  = detUncerts
      end if

      do i = 1, size(raDeg)
        if (detX(i) * detX(i) + detY(i) * detY(i) > fovRadiusDet2) then
          srcNumStr = intToStr(int(i))
          write(errstr, '(a)') 'The calculated det XY position &
            &for source '//trim(srcNumStr)//' does not fall within the &
            &field of view.'

          call warning('notInFov')
          call setToNull(detXcol, i - 1)
          call setToNull(detYcol, i - 1)
          if (withUncertCol) call setToNull(detUncertsCol, i - 1)
        end if
      end do

      call release(set)

    case('user')
      outputStr = '' ! fills it with blanks.

      if (detX(1) * detX(1) + detY(1) * detY(1) <= fovRadiusDet2) then
        if (withHeader) then
          write(*, '(a)') '# Instrument: '//trim(instId(instrumentId))
          write(*, '(a)') '# Coord sytem of output is DETXY (CAMCOORD2 &
            &but in units of 0.05 arcsec).'

          write(*, '(a, f10.6, a)') '# Source RA = ', raDeg(1), ' deg.'
          write(*, '(a, f10.6, a)') '# Source dec = ', decDeg(1), ' deg.'
        end if

        if (withUncertRadius) then
          if (withHeader) then
            write(*, '(a)') '#'
            write(*, '(a)') '# detX       detY       det err'
!                           '+xxxxx.x   +xxxxx.x     xxxxx.x'
          end if
          write(*, '(f8.1, a3, f8.1, a5, f7.1)') detX(1), '   ', detY(1)&
            , '     ', detUncerts(1)

        else
          if (withHeader) then
            write(*, '(a)') '#'
            write(*, '(a)') '# detX       detY'
!                           '+xxxxx.x   +xxxxx.x'
          end if
          write(*, '(f8.1, a3, f8.1)') detX(1), '   ', detY(1)
        end if

      else
call error('notInFov','The calculated det XY position does not fall &
&within the field of view.')
      end if

      print *, outputStr
!*** send it as a message???

    case default
call error('badDataStyle', "The value of the --datastyle parameter &
&was not recognised.")
    end select

  end subroutine detOutput

  !---------------------------------------------------------------------
  subroutine rawOutput(raDeg, decDeg, rawX, rawY, rawXUncerts, rawYUncerts&
  , offChip, ccdNr, instrumentId)
  !---------------------------------------------------------------------

    implicit none

    real(double),   intent(in) :: raDeg(:),&
                                  decDeg(size(raDeg))
    real(single), intent(in) :: rawX(size(raDeg)),&
                                  rawY(size(raDeg)),&
                                  rawXUncerts(size(raDeg)),&
                                  rawYUncerts(size(raDeg))
    logical(bool),  intent(in) :: offChip(:)
    integer(int8),  intent(in) :: ccdNr(size(raDeg))
    integer(int32), intent(in) :: instrumentId

    type(DataSetT) :: set
    type(TableT)   :: tab
    type(ColumnT)  :: rawXcol, rawYcol, ccdNrCol, rawXUncertsCol,&
      rawYUncertsCol, flagCol
    real(single), dimension(:), pointer ::&
      dalRawX, dalRawY, dalRawXuncerts, dalRawYuncerts
    integer(int8), dimension(:), pointer :: dalCcdNr
    integer(int32), dimension(:), pointer :: dalFlag
    character(37) :: outputStr
!    integer(int16) :: i
!    character(12) :: srcNumStr
character(7) :: onChipStr

    !-------------------------------------------------------------------
    ! Start

    call message(AppMsg, Verbose, 'Writing output data.')

    select case(dataStyle)
    case('set')
      if (withOutSet) then
        call splitSetTabName(outSetTabName, outSetName, outTabName)
        ! in ssclib/dal_aux

if (outTabName .eq. '') call error('emptyTableName',"Parameter --outtab &
&should be given in the form dataSetName:tableName")

        set = dataSet(outSetName, CREATE)
        call setAttribute(set, 'INSTRUME', trim(instId(instrumentId)))
!*** put in DATE-TIME and wcs and *_PNT kwds too??
        tab = addTable(set, outTabName, size(rawX))

      else
        set = dataSet(inSetName, MODIFY)
        tab = table(set, inTabName, MODIFY)
      end if

      rawXcol  = addOrOpenColumn(tab, 'RAWX',  REAL32, 'CCD pixels')
      rawYcol  = addOrOpenColumn(tab, 'RAWY',  REAL32, 'CCD pixels')
      ccdNrCol = addOrOpenColumn(tab, 'CCDNR', INTEGER8)
      flagCol  = addOrOpenColumn(tab, 'FLAG',  INTEGER32)
      ! addOrOpenColumn is in ssclib/dal_aux.

      if (withUncertCol) then
        rawXUncertsCol  = addOrOpenColumn(tab, 'RAWX_ERR', REAL32&
          , 'CCD pixels')
        rawYUncertsCol  = addOrOpenColumn(tab, 'RAWY_ERR', REAL32&
          , 'CCD pixels')
      end if

      dalRawX  => real32Data(rawXcol)
      dalRawY  => real32Data(rawYcol)
      dalCcdNr => int8Data(ccdNrCol)
      dalFlag  => int32Data(flagCol)
      dalRawX  = rawX
      dalRawY  = rawY
      dalCcdNr = ccdNr
      dalFlag  = 0

      where(offChip) dalFlag = ibset(dalFlag, OFF_CHIP_BIT)

      if (withUncertCol) then
        dalRawXuncerts  => real32Data(rawXUncertsCol)
        dalRawYuncerts  => real32Data(rawYUncertsCol)
        dalRawXuncerts  = rawXUncerts
        dalRawYuncerts  = rawYUncerts
      end if

      call release(set)

    case('user')
      outputStr = '' ! fills it with blanks.

      if (withHeader) then
        write(*, '(a)') '# Instrument: '//trim(instId(instrumentId))
        write(*, '(a)') '# Coord sytem of output is RAWXY (PIXCOORD1).'
        write(*, '(a, f10.6, a)') '# Source RA = ', raDeg(1), ' deg.'
        write(*, '(a, f10.6, a)') '# Source dec = ', decDeg(1), ' deg.'
      end if

      if (offChip(1)) then
        onChipStr = '      F'
call warning('notOnChip', 'The calculated raw XY position does not fall &
&on any ccd. The position relative to the nearest CCD has been printed.')
      else
        onChipStr = '      T'
      end if

      if (withUncertRadius) then
        if (withHeader) then
          write(*, '(a)') '#'
!          write(*, '(a)') '# rawX      rawY    ccd      X err       Y err   On chip?'
!!                         '+xxx.xx   +xxx.xx   xx   x.xxxE+xx   x.xxxE+xx     x'
write(*, '(a)') '# rawX      rawY    ccd   On chip?      X err       Y err'
!                +xxx.xx   +xxx.xx   xx      x       x.xxxE+xx   x.xxxE+xx
        end if
!        write(*, '(i3, a8, i3, a5, i2, a3, i3, a7, i3)') rawX(1)&
!write(*, '(f7.2, a3, f7.2, a3, i2, a3, e9.3, a3, e9.3, a7)') rawX(1)&
!          , '   ', rawY(1), '   ', ccdNr(1), '   '&
!          , rawXUncerts(1), '   ', rawYUncerts(1), onChipStr
write(*, '(f7.2, a3, f7.2, a3, i2, a7, a7, e9.3, a3, e9.3)') rawX(1)&
, '   ', rawY(1), '   ', ccdNr(1), onChipStr, '       '&
, rawXUncerts(1), '   ', rawYUncerts(1)

      else
        if (withHeader) then
          write(*, '(a)') '#'
          write(*, '(a)') '# rawX      rawY    ccd   On chip?'
!                         '+xxx.xx   +xxx.xx   xx      x'
        end if
!        write(*, '(i3, a8, i3, a5, i2)') rawX(1), '        '&
write(*, '(f7.2, a3, f7.2, a3, i2, a7)') rawX(1), '   '&
          , rawY(1), '   ', ccdNr(1), onChipStr

      end if

      print *, outputStr
!*** send it as a message???

    case default
call error('badDataStyle', "The value of the --datastyle parameter &
&was not recognised.")
    end select

  end subroutine rawOutput

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module esky2det_mod
